#include<stdio.h>
#include<math.h>
#include <stdlib.h>
#include "A_comp.h"
#include "allocate.h"

#define LEN 512


void A_comp_ij(
 int im_row, int im_col,
 struct sino_info *sino,
 struct geom_info *geom,
 double **pix_prof,
 struct column *sp_col);

void pix_prof_comp(
		   struct sino_info *sino,
		   struct geom_info *geom,
		   double ***pix_prof)
{
  
  int	i,j;
  double pi,ang,d1,d2,t,t_1,t_2,t_3,t_4,maxval,scale,rc;
  
  *pix_prof = (double **)get_img(LEN,sino->n_theta,sizeof(double));
  
  pi = 4.0 * atan(1.0);
  
  /* Fill profiles of pixels from all angles
     Each profile assumed 2 pixels wide
     This is used to speed calculation (too many to store) of the
     entries of the projection matrix.
     Values are scaled by "scale"
  */
  /* WATCH THIS; ONLY FOR SQUARE PIXELS NOW 	*/
  
  scale = geom->delta;
  
  rc = sin((double)(pi/4.0));
  for(i = 0; i < sino->n_theta; i++) {
    ang = sino->theta_0+i*sino->delta_theta;
    
    while(ang >= pi/2.0)
      ang -= pi/2.0;
    
    while(ang < 0.0)
      ang += pi/2.0;
    
    if(ang <= pi/4.0)
      maxval = scale/cos((double) ang);
    else
      maxval = scale/cos((double) (pi/2.0-ang));
    
    d1 = rc * cos((double) (pi/4.0-ang));
    d2 = rc*fabs((double) (sin((double)(pi/4.0-ang))));
    t_1 = 1.0-d1;
    t_2 = 1.0-d2;
    t_3 = 1.0+d2;
    t_4 = 1.0+d1;
    for(j = 0; j < LEN; j++) {
      t = 2.0*j / (double)LEN;
      if(t<=t_1 || t>=t_4)
	(*pix_prof)[i][j] = 0.0;
      else if(t>t_3)
	(*pix_prof)[i][j] = maxval*(t_4-t)/(t_4-t_3);
      else if(t>=t_2)
	(*pix_prof)[i][j] = maxval;
      else
	(*pix_prof)[i][j] = maxval*(t-t_1)/(t_2-t_1);
    }
  }

} 




void A_comp(
	    int im_row, int im_col,
	    struct sino_info *sino,
	    struct geom_info *geom,
	    double **pix_prof,
	    struct column *sp_col)
{
  
#ifdef STORE_A_MATRIX
  /* precompute entire A-matrix at first call, then 
     just return precomputed cols in subsequent calls */
  
  static int first=1;
  static struct column sp_col_sgl;
  static struct column_flt **sp_col_arr,*spc_f;
  int i,j,r,nr;
  
  
  if (first==1)
    {
      first=0;
      
      printf("Initializing A-matrix...");fflush(stdout);
      sp_col_sgl.val = (double *) get_spc(sino->n_t*sino->n_theta, sizeof(double));
      sp_col_sgl.row = (int *) get_spc(sino->n_t*sino->n_theta, sizeof(int));
      
      fflush(stdout);fflush(stderr);
      
      sp_col_sgl.multiplier = 1.0; 
      
      sp_col_arr=(struct column_flt **)multialloc(sizeof(struct column_flt),2,geom->n_y,geom->n_x);
      
      
      for (i=0;i<geom->n_y;i++)
	for (j=0;j<geom->n_x;j++)
	  {
	    A_comp_ij(i,j,sino,geom,pix_prof,&sp_col_sgl);
	    
      	    sp_col_arr[i][j].val=(float*)get_spc(sp_col_sgl.n_row, sizeof(float));
	    sp_col_arr[i][j].row = (int*)get_spc(sp_col_sgl.n_row, sizeof(int));
	    sp_col_arr[i][j].n_row = sp_col_sgl.n_row;
	    sp_col_arr[i][j].multiplier = sp_col_sgl.multiplier;
	    for (r=0;r<sp_col_sgl.n_row;r++)
	      {
		sp_col_arr[i][j].val[r] = (float)sp_col_sgl.val[r];
		sp_col_arr[i][j].row[r] = sp_col_sgl.row[r];
	      }
	    
	  }
      free((void*)sp_col_sgl.val);
      free((void*)sp_col_sgl.row);
      fflush(stderr);
      printf("done\n");fflush(stdout);
    }
  
  spc_f=&sp_col_arr[im_row][im_col];
  sp_col->n_row = nr = spc_f->n_row;
  sp_col->multiplier = spc_f->multiplier;
  
  for (r=0;r<nr;r++)
    {
      sp_col->val[r]=(double)spc_f->val[r];
      sp_col->row[r]=spc_f->row[r];
    }
  
#else
  /* compute A_*j on the fly */
  A_comp_ij(im_row,im_col,sino,geom,pix_prof,sp_col);
#endif
}



void A_comp_ij(
	       int im_row, int im_col,
	       struct sino_info *sino,
	       struct geom_info *geom,
	       double **pix_prof,
	       struct column *sp_col)
  
{
  int ind,ind_min,ind_max,pr;
  int prof_ind,pind,i,proj_count;
  
  double	Aval,t_min,t_max,ang,x,y;
  
  /* WATCH THIS; ONLY FOR SQUARE PIXELS NOW 	*/
  
  
  y = geom->y_0 + im_row*geom->delta;
  x = geom->x_0 + im_col*geom->delta;
  
  proj_count = 0;
  for(pr=0; pr < sino->n_theta; pr++)
    {
      ang = sino->theta_0 + pr*(sino->delta_theta);
      pind = pr*sino->n_t;
      
      /* Need profile to contain 2 pixel widths       */
      t_min = y*cos((double) ang) - x*sin((double) ang) - geom->delta;
      t_max = t_min + 2.0*geom->delta;
      if(t_max < sino->t_0) /* This also prevents over-reach (with rounding of
			       negative numbers)	*/
	continue;
      
      
      
      /* Changed 10/2000 (TF): more general, old implementation assumed
	 that geom->dia=sino->n_t*sino->delta_t; */ 
      ind_min = ceil((t_min-sino->t_0)/sino->delta_t);
      ind_max = (t_max-sino->t_0)/sino->delta_t;
      
      /* Fix this 4/91 to prevent over-reach at ends  */
      ind_min = (ind_min<0) ? 0:ind_min;
      ind_max = (ind_max>=sino->n_t) ? sino->n_t-1:ind_max;
      for(i = ind_min; i <=ind_max; i++) 
	{
	  
	  ind = pind+i;
	  prof_ind = LEN*(sino->t_0+i*sino->delta_t-t_min)/(2.0*geom->delta);
	  
	  if(prof_ind >= LEN || prof_ind <0){
	    if(prof_ind == LEN)
	      prof_ind = LEN-1;
	    else if (prof_ind == -1)
	      prof_ind = 0;
	    else {
	      fprintf(stderr,"\nExiting Program: input parameters inconsistant\n");
	      exit(-1);
	    }
	  }
	  
	  if((Aval=pix_prof[pr][prof_ind]) > 0.0) {
 	    sp_col->val[proj_count] = Aval;
 	    sp_col->row[proj_count] = ind;/* Index for ray    */
	    proj_count++;
	  }
	}
    } 
  
  sp_col->n_row = proj_count;
  
  
}
