#include <stdio.h>
#include <math.h>
#include "assort.h"
#include "allocate.h"



double sq(double x)
{
	return(x*x);
}

double energy_logcosh_toroid(float **img, int N, int M, float sigma)
{
    int i,j;
	int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;




    for(j=0;j<M;j++)
	    for(i=0;i<N;i++)
        {
            ai = (i+1)%N;
			aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            sum += logcosh((img[i][j] - img[ai][j])/sigma);
            sum += logcosh((img[i][j] - img[i][aj])/sigma);
            sumd += logcosh((img[i][j] - img[ai][sj])/sigma);
            sumd += logcosh((img[i][j] - img[ai][aj])/sigma);
        }
	
	sum = sum/(4.0+2.0*sqrt(2.0));
	sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}

double energy_logcosh(float **img, int N, int M, float sigma)
{
    int i,j;
    double sum = 0.0, sumd = 0.0;


	j=0;
    for(i=0;i<N-1;i++) 
    {  
        sum += logcosh((img[i][j] - img[i+1][j])/sigma);   
        sum += logcosh((img[i][j] - img[i][j+1])/sigma);   
        sumd += logcosh((img[i][j] - img[i+1][j+1])/sigma);
    }  

	j = M-1;
    for(i=0;i<N-1;i++) 
    {  
        sum += logcosh((img[i][j] - img[i+1][j])/sigma);   
        sumd += logcosh((img[i][j] - img[i+1][j-1])/sigma);
    }  

	i = N-1;
    for(j=0;j<M-1;j++) 
        sum += logcosh((img[i][j] - img[i][j+1])/sigma);   


    for(j=1;j<M-1;j++)
	    for(i=0;i<N-1;i++)
        {
            sum += logcosh((img[i][j] - img[i+1][j])/sigma);
            sum += logcosh((img[i][j] - img[i][j+1])/sigma);
            sumd += logcosh((img[i][j] - img[i+1][j-1])/sigma);
            sumd += logcosh((img[i][j] - img[i+1][j+1])/sigma);
        }
	
	sum = sum/(4.0+2.0*sqrt(2.0));
	sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}

double logcosh(double x)
{
    x = fabs(x);
    if(x > 10.0)
        return(x - log(2));
    else
        return(log(cosh(x)));
}


double energy8_toroid(float **img, int N, int M, float p)
{
    int i,j;
	int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;


    for(j=0;j<M;j++)
	    for(i=0;i<N;i++)
        {
            ai = (i+1)%N;
			aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            sum += pow(fabs(img[i][j] - img[ai][j]),p);
            sum += pow(fabs(img[i][j] - img[i][aj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][sj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][aj]),p);
        }
	
	sum = sum/(4.0+2.0*sqrt(2.0));
	sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}
double energy8_toroid_dbl(double **img, int N, int M, float p)
{
    int i,j;
        int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;


    for(j=0;j<M;j++)
            for(i=0;i<N;i++)
        {
            ai = (i+1)%N;
                        aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            sum += pow(fabs(img[i][j] - img[ai][j]),p);
            sum += pow(fabs(img[i][j] - img[i][aj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][sj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][aj]),p);
        }

        sum = sum/(4.0+2.0*sqrt(2.0));
        sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}


double dp_energy8_toroid(float **img, int N, int M, float p)
{
    int i,j;
	int aj,sj,ai;
    double sum = 0.0, sumd = 0.0,temp;


    for(j=0;j<M;j++)
	    for(i=0;i<N;i++)
        {
            ai = (i+1)%N;
			aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

			temp = fabs(img[i][j] - img[ai][j]);
			if(temp != 0.0)
              sum += log(temp)*pow(temp,p);

			temp = fabs(img[i][j] - img[i][aj]);
			if(temp != 0.0)
              sum += log(temp)*pow(temp,p);

			temp = fabs(img[i][j] - img[ai][sj]);
			if(temp != 0.0)
              sumd += log(temp)*pow(temp,p);

			temp = fabs(img[i][j] - img[ai][aj]);
			if(temp != 0.0)
              sumd += log(temp)*pow(temp,p);
        }
	
	sum = sum/(4.0+2.0*sqrt(2.0));
	sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}

double dcostp(float **img, int N, int M, float p)
{
    int i,j;
    int aj,sj,ai;
    double sum = 0.0, sumd = 0.0,temp,tempp;
    double dsum = 0.0, dsumd = 0.0;
	double T,du;


/*
    for(j=1;j<M-1;j++)
        for(i=1;i<N-1;i++)
		*/
    for(j=0;j<M;j++)
        for(i=0;i<N;i++)
        {
            ai = (i+1)%N;
            aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            temp = fabs(img[i][j] - img[ai][j]);
			tempp = pow(temp,p);
            sum += tempp;
            if(temp != 0.0)
              dsum += log(temp)*tempp;

            temp = fabs(img[i][j] - img[i][aj]);
			tempp = pow(temp,p);
            sum += tempp;
            if(temp != 0.0)
              dsum += log(temp)*tempp;

            temp = fabs(img[i][j] - img[ai][sj]);
			tempp = pow(temp,p);
            sumd += tempp;
            if(temp != 0.0)
              dsumd += log(temp)*tempp;

            temp = fabs(img[i][j] - img[ai][aj]);
			tempp = pow(temp,p);
            sumd += tempp;
            if(temp != 0.0)
              dsumd += log(temp)*tempp;
        }

/*
    N = N-2;
    M = M-2;
	*/

    sum = sum/(4.0+2.0*sqrt(2.0));
    sumd = sumd/(4.0+4.0*sqrt(2.0));
	T = p*(sum+sumd)/(N*M);

    dsum = dsum/(4.0+2.0*sqrt(2.0));
    dsumd = dsumd/(4.0+4.0*sqrt(2.0));
	du = (dsum+dsumd)/(N*M);


    return(-log(T)/(p*p) + du/T );
}

double dcostp_sup(float **img, int **mask, int N, int M, int Nsup, float p)
{
    int i,j;
    int aj,sj,ai;
    double sum = 0.0, sumd = 0.0,temp,tempp;
    double dsum = 0.0, dsumd = 0.0;
    double T,du;


    for(j=0;j<M;j++)
        for(i=0;i<N;i++)
		  if( mask[i][j] )
       	  {
            ai = (i+1)%N;
            aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            temp = fabs(img[i][j] - img[ai][j]);
            tempp = pow(temp,p);
            sum += tempp;
            if(temp != 0.0)
              dsum += log(temp)*tempp;

            temp = fabs(img[i][j] - img[i][aj]);
            tempp = pow(temp,p);
            sum += tempp;
            if(temp != 0.0)
              dsum += log(temp)*tempp;

            temp = fabs(img[i][j] - img[ai][sj]);
            tempp = pow(temp,p);
            sumd += tempp;
            if(temp != 0.0)
              dsumd += log(temp)*tempp;

            temp = fabs(img[i][j] - img[ai][aj]);
            tempp = pow(temp,p);
            sumd += tempp;
            if(temp != 0.0)
              dsumd += log(temp)*tempp;
          }


    sum = sum/(4.0+2.0*sqrt(2.0));
    sumd = sumd/(4.0+4.0*sqrt(2.0));
    T = p*(sum+sumd)/Nsup;

    dsum = dsum/(4.0+2.0*sqrt(2.0));
    dsumd = dsumd/(4.0+4.0*sqrt(2.0));
    du = (dsum+dsumd)/Nsup;


    return(-log(T)/(p*p) + du/T );
}



double energy8_sup_dbl(double **img, int **mask, int N, int M, double p)
{
    int i,j;
    int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;


    for(j=0;j<M;j++)
        for(i=0;i<N;i++)
		  if( mask[i][j] )
          {
            ai = (i+1)%N;
            aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            sum += pow(fabs(img[i][j] - img[ai][j]),p);
            sum += pow(fabs(img[i][j] - img[i][aj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][sj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][aj]),p);
          }

    sum = sum/(4.0+2.0*sqrt(2.0));
    sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}


double energy8_sup(float **img, int **mask, int N, int M, float p)
{
    int i,j;
    int aj,sj,ai;
    double sum = 0.0, sumd = 0.0;


    for(j=0;j<M;j++)
        for(i=0;i<N;i++)
                  if( mask[i][j] )
          {
            ai = (i+1)%N;
            aj = (j+1)%M;
            sj = j-1;
            if (sj < 0)
                sj = M-1;

            sum += pow(fabs(img[i][j] - img[ai][j]),p);
            sum += pow(fabs(img[i][j] - img[i][aj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][sj]),p);
            sumd += pow(fabs(img[i][j] - img[ai][aj]),p);
          }

    sum = sum/(4.0+2.0*sqrt(2.0));
    sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}


double energy8_bod(float **img, int N, int M, float p)
{
    int i,j;
    double sum = 0.0, sumd = 0.0;


    for(j=1;j<M-1;j++)
	    for(i=1;i<N-1;i++)
        {
            sum += pow(fabs(img[i][j] - img[i+1][j]),p);
            sum += pow(fabs(img[i][j] - img[i][j+1]),p);
            sumd += pow(fabs(img[i][j] - img[i+1][j-1]),p);
            sumd += pow(fabs(img[i][j] - img[i+1][j+1]),p);
        }
	
	sum = sum/(4.0+2.0*sqrt(2.0));
	sumd = sumd/(4.0+4.0*sqrt(2.0));

    return((sum+sumd));
}
	
double energy4_prec(float **img, int N, int M, float p)
{
    int i,j;
    double sum = 0.0;


    for(i=0;i<N-1;i++)
        for(j=0;j<M-1;j++)
        {
            sum += pow(fabs(img[i][j] - img[i+1][j]),p);
            sum += pow(fabs(img[i][j] - img[i][j+1]),p);
        }

    return( sum/4.0 );
}



float max(float **img,int N, int M)
{
	int i,j;
	float max;

	max = img[0][0];

	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			if (img[i][j] > max)
				max = img[i][j];
	
	return(max);
}

float min(float **img,int N, int M)
{
	int i,j;
	float min;

	min = img[0][0];
	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			if ( img[i][j] < min )
				min = img[i][j];
	
	return(min);
}

double mean(float **img, int N, int M)
{
	int i,j;
	double sum=0.0;

	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			sum = sum + img[i][j];
	
	return(sum/(N*M));
}

double mean_dbl(double **img, int N, int M)
{
        int i,j;
        double sum=0.0;

        for (i=0;i<N;i++)
                for(j=0;j<M;j++)
                        sum = sum + img[i][j];

        return(sum/(N*M));
}


void absolute(float **img, int N, int M)
{
	int i,j;


	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			img[i][j] = fabs(img[i][j]);
}

void sqrt_img(float **img, int N, int M)
{
	int i,j;


	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			img[i][j] = sqrt(img[i][j]);
}

void offset(float **img, int N, int M, float offset)
{
	int i,j;


	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			img[i][j] = img[i][j] + offset;
	
}

void scale_img(float **img, int N, int M, float scale, float max)
{
	int i,j;


	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			img[i][j] = img[i][j]*max/scale;

}

void copy_ch_fl(unsigned char **img1, float **img2, int N, int M)
{
	int i,j;

	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
			img2[i][j] = (float) img1[i][j];
}

void copy_fl_ch(float **img1, unsigned char **img2, int N, int M)
{
	int i,j,temp;

	for (i=0;i<N;i++)
		for(j=0;j<M;j++)
		{
			temp = (int)(img1[i][j]+0.5);
			if (temp < 0)
				temp = 0;
			else if (temp > 255)
				temp = 255;
			img2[i][j] = (unsigned char) temp;
		}
}
