#include "fr_icm.h"


double ICMstep(struct ICMdata *ICMdata,struct all_data *data,double *S,int i,int j,struct geom_info *geom,struct column *sp_col)
{	  
  double mm;
  

  /* Compute D1 and D2 for quadratic approx. of liklihood ftn */
  compute_derivatives(ICMdata, data, sp_col,S);
  
  /* compute ML extimate */
  ICMdata->ML = data->X[i][j] - ICMdata->D1/ICMdata->D2;

  /* get neighbors of (i,j) */
  get_neighborhood(ICMdata,data,geom,i,j);

  /* compute ICM-update */
  mm = updateX(ICMdata);
 
  return(mm);
}


double updateX(struct ICMdata *ICMdata)
{
  double mm;
  double low,high;
  int status;

  /* calculate limits for half-interval search */
  minmax(&low, &high,ICMdata->ML,ICMdata->neigh.NX,ICMdata->neigh.DX);
  
 

  
  /* ICM */
  /* root derivative of approx. energy */
  /* this yields mm = mode of distribution */
  mm = solve(diff_cost, low, high, TOL, &status,(void*)(ICMdata));
  if (status)
    {
      if (cost(low,ICMdata) < cost(high,ICMdata))
	mm = low;
      else
	mm = high;
    }
  
  return(mm);
}




void get_neighborhood(struct ICMdata *ICMdata,struct all_data *data,struct geom_info *geom,int i,int j)
{
  int ai,aj,si,sj,N,M;
  double **X,**Y;
  

  X=data->X;
  Y=data->Y;
  N = geom->n_y;
  M = geom->n_x;


  ai = (i+1)%N;
  aj = (j+1)%M;
  si = i-1;
  if (si < 0)
    si = N-1;
  sj = j-1;
  if (sj < 0)
    sj = M-1;
	      
  ICMdata->neigh.NX[0] = X[i][aj];
  ICMdata->neigh.NX[1] = X[i][sj];
  ICMdata->neigh.NX[2] = X[ai][j];
  ICMdata->neigh.NX[3] = X[si][j];
  
  ICMdata->neigh.DX[0] = X[ai][aj];
  ICMdata->neigh.DX[1] = X[ai][sj];
  ICMdata->neigh.DX[2] = X[si][aj];
  ICMdata->neigh.DX[3] = X[si][sj];
  ICMdata->CX = X[i][j];

}




void compute_derivatives(struct ICMdata *ICMdata, struct all_data *data, struct column *sp_col,double *S)
{
  double D1,D2;
  int FLAG;
  int t,pro;
  double lambda=0.0,p_hat=0.0;

  D1 = 0.0;
  D2 = 0.0;
  FLAG = 0;
  for(t=0;t<sp_col->n_row;t++)
    {
      pro = sp_col->row[t];
      if (S[pro] >= 0.0)
	{
	  p_hat = exp(-S[pro]);
	  lambda = data->Yt[0][pro]*p_hat;
	  D1 += sp_col->val[t]*(data->Y[0][pro] - lambda);
	  D2 += sp_col->val[t]*sp_col->val[t]*lambda;
	}
    }

  ICMdata->D1=D1;
  ICMdata->D2=D2;
}	




void updateS(double *S, struct column *sp_col, double diff)
{
  int t,pro;
  
  for(t=0;t<sp_col->n_row;t++)
    {
      pro = sp_col->row[t];
      S[pro] = S[pro] + diff*sp_col->val[t];
    }
}



/* based on value of iter, uses different pixel order and returns locations i,j*/
/* for pixel number s*/
void imgcoord(int *out_i,int *out_j,int s,int iter,int N,int M)
{  
  int i,j,ii,jj,MM,NN,itemp;

  if((iter/4)%2 != 0)
    {
      NN = M;
      MM = N;
    }
  else
    {
      NN = N;
      MM = M;
    }
  
  ii=s/MM;
  jj=s-ii*MM;
  
  
  if((iter-1)%4 == 0)
    {
      i = NN-1-ii;
      j = MM-1-jj;
    }
  else 
    if((iter-2)%4 == 0)
      {
	i = NN-1-ii;
	j = jj;
      }
    else 
      if((iter-3)%4 == 0)
	{
	  i = ii;
	  j = MM-1-jj;
	}
      else
	{
	  i = ii;
	  j = jj;
	}
  
  if((iter/4)%2 != 0)
    {
      itemp = i;
      i = j;
      j = itemp;
    }
  
  *out_j=j;
  *out_i=i;

}
 





void minmax(double *low, double *high, double ML, double *NX, double *DX)
{
  int i;
  
  *low = ML;
  *high = ML;
  
  for(i=0;i<4;i++)
    {
      
      if( *low > NX[i])
	*low = NX[i];
      else if(*high < NX[i])
	*high = NX[i];
      
      if( *low > DX[i])
	*low = DX[i];
      else if(*high < DX[i])
	*high = DX[i];
      
    }
}




double diff_cost(double x,void *ICMdata)
{
    double prior=0.0, dprior=0.0;
    int i;
    struct ICMdata *par;
    
    par=(struct ICMdata*)ICMdata;
    
    
    
    for(i=0;i<4;i++)
      {
	prior += pow(fabs(x - par->neigh.NX[i]),par->P)*SG(x,par->neigh.NX[i]);
	dprior += pow(fabs(x - par->neigh.DX[i]),par->P)*SG(x,par->neigh.DX[i]);
      }
    prior  = prior*WN;
    dprior  = dprior*WD;
    
    return( par->D1 + par->D2*(x - par->CX) + (prior+dprior)/par->SX );
}



double cost(double x,struct ICMdata *par)
{
  double temp, prior=0.0, dprior=0.0;
  int i;
  
  temp = x-par->CX;
  
  for(i=0;i<4;i++)
    {
      prior += pow(fabs(x - par->neigh.NX[i]),par->P+1);
      dprior += pow(fabs(x - par->neigh.DX[i]),par->P+1);
    }
  prior  = prior*WN;
  dprior  = dprior*WD;
  
  return( par->D1*temp + par->D2*temp*temp/2.0 + (prior + dprior )/((par->P+1)*par->SX) );
}



void  f_pro(double *S,struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col)
{
  int i,j,t;
  int M,N;


  N = geom->n_y;
  M = geom->n_x;
 
  for(i=0;i<data->sino.n_t*data->sino.n_theta;i++)
    S[i] = 0.0;
  
  for(i=0;i<N;i++)
    for(j=0;j<M;j++)
      {
	A_comp(i,j,&data->sino,geom,profiles,sp_col);
        for(t=0;t<sp_col->n_row; t++)
	  S[sp_col->row[t]] = S[sp_col->row[t]] + sp_col->val[t]*data->X[i][j];
      }
}




