#include "fr_paramest.h"

double delta_cost(double x,struct ICMdata *ICMdata, struct column *sp_col, struct all_data *data, double *S) ;
double H_Metropolis_new(struct ICMdata *ICMdata,double mm,double *S,struct column *sp_col,struct all_data *data);

double parameter_estimation(struct all_data *data, struct geom_info *geom, double **profiles, struct cmdline_params *cmdline, struct column *sp_col, double P)
{
  double *Xinit;
  double *S; /* Ax */
 
  double SX;
  int i,N,M;
 
  N = geom->n_y;
  M = geom->n_x;
  srandom2(1);
   
  /* save initial X */
  Xinit = (double *) get_spc(N*M, sizeof(double));
  for( i=0;i<N*M;i++ )
    Xinit[i] = data->X[0][i];

  /* Calulate S = Ax */
  S = (double *) get_spc(data->sino.n_t*data->sino.n_theta, sizeof(double));
  f_pro(S,data,geom,profiles,sp_col);

  /* initialize sigma */
  SX = energy8_sup_dbl(data->X, data->mask, N, M,P+1)/data->Nsup;
  if (SX == 0.0)
    SX = mean_dbl(data->X,N,M);
  printf("Sigma initialized to %f\n",pow(SX,1.0/cmdline->p));
  printf("Starting estimation of sigma ...\n");
  
  /* Estimate sigma using ICM/Hastings-Metropolis */
  SX=relax_param(data,S, geom, profiles, cmdline,sp_col,P,SX);

  /* set X back to Xinit */
  for( i=0;i<N*M;i++ )
    data->X[0][i] = Xinit[i];

  free((void*)S);

  return(SX);
}



double enter_parameter(struct all_data *data, struct geom_info *geom,struct cmdline_params *cmdline, double P)
{
  double SX;
  int N,M;
 
  N = geom->n_y;
  M = geom->n_x;
   
  SX = energy8_sup_dbl(data->X, data->mask, N, M,P+1)/data->Nsup;
  printf("You have set the no. of iterations for parameter estimation to be 0\n");
  printf("Sigma computed from the initial condition is %f\n",pow(SX,1.0/cmdline->p));
  printf("Enter desired value for sigma: \n");
  scanf("%lf",&SX);
  SX = pow(SX,cmdline->p);

  return(SX);

} 

double relax_param(struct all_data *data,double *S,struct geom_info *geom, double **profiles, struct cmdline_params *cmdline,struct column *sp_col,double P,double SX)
{
  
  int i,j,s,EMiter;
  int N,M;
  int count=0; /* count for sigma proj. */

  double **X,**Y;
  double diff,mm;
  double SXold;
  double x[PN],y[PN];
  double proj_sigma=0;

  double MT_SX;
  int nb_MT_iter=1;
  int MTiter;
  struct ICMdata ICMdata;

  /*FILE *fp;
  FILE *fp2;
  char fname[50];*/

  X=data->X;
  Y=data->Y;
  N = geom->n_y;
  M = geom->n_x;

  /* ICM data */
  ICMdata.SX=SX;
  ICMdata.P=P;

  /*cmdline->no_iter_param=1;
    ICMdata.SX=pow(0.08,P+1);*/

  for(EMiter=0;EMiter<cmdline->no_iter_param;EMiter++)
    {
      /*sprintf(fname,"tmp/em%d",EMiter);*/
      /*fp=fopen(fname,"wa");*/
      for (MTiter=0;MTiter<nb_MT_iter;MTiter++)
	{
	  for (s=0;s<N*M;s++) /* loop over pixel locations */
	    {
	      
	      /* determine i,j image location, depending on iter go in*/
	      /* row or column order and also start from different corners */
	      imgcoord(&i,&j,s,MTiter,N,M);
	      
	      /* compute column of A matrix for loc. i,j */
	      A_comp(i,j,&data->sino,geom,profiles,sp_col);
	      
	      /* if more than 0 projections intersect location i,j */
	      if(sp_col->n_row)
		{
		  mm=ICMstep(&ICMdata,data,S,i,j,geom,sp_col);
		  
		  /*use ICM update as mean */
		  /* and sample from Gaussian (/exp) distrib. */ 
		  X[i][j]=H_Metropolis_new(&ICMdata,mm,S,sp_col,data);
		  
		 
		  
		  /* If pixel-value was changed, update S=Ax */
		  diff = X[i][j] - ICMdata.CX;
		  if(diff)
		    updateS(S,sp_col,diff);
		  
		  
		} /* end if sp_col */
	    } /* end for s */
	  
	  /*calculate new sigma */ 
	  /* and extrapolate */
	  MT_SX=energy8_sup_dbl(X, data->mask, N, M,P+1)/data->Nsup;
	  printf("MT_iter:%d ICMdata.SX=%f SX=%f \n",MTiter,pow(ICMdata.SX,1.0/(P+1)),pow(MT_SX,1.0/(P+1)));
	  /*fprintf(fp,"%f\n",pow(MT_SX,1.0/(P+1)));fflush(fp);*/

	  /* write metrop iter */
	  /*sprintf(fname,"tmp/x%d",MTiter);
	    fp2=fopen(fname,"wa");
	    for (i=0;i<N;i++)
	    {
	    for (j=0;j<M;j++)
	    fprintf(fp2,"%f ",X[i][j]);
	    fprintf(fp2,"\n");
	    }
	    fclose(fp2);*/
		  

	}
      /*fclose(fp);*/

      SXold=ICMdata.SX;
      ICMdata.SX = energy8_sup_dbl(X, data->mask, N, M,P+1)/data->Nsup;
      proj_sigma=extrapolate_sigma(x, y,ICMdata.SX,SXold,data, P, count);
      count++;
      printf("%d. sigma %f, extrapolated sigma %f\n",EMiter+1,pow(ICMdata.SX,1.0/(P+1)),pow(proj_sigma,1.0/(P+1)));
      
    } /* end for iter */
  
 
  if(proj_sigma > 0.0 && cmdline->no_iter_param >= 3)
    ICMdata.SX = proj_sigma;
  printf("Estimated value of sigma %f\n",pow(ICMdata.SX,1.0/(P+1)));
  
  
  return(ICMdata.SX);
}	


double extrapolate_sigma(double *x, double *y, double SX, double SXold, struct all_data *data, double P, int count)	  
{  
  int pn;
  double proj_sigma;

  x[count%PN] = (P+1)*SXold;
  y[count%PN] = (P+1)*SX - x[count%PN];
  
  
  if(count+1 > PN)
    pn = PN;
  else
    pn = count+1;
  if(pn > 1)
    proj_sigma = proj(x,y,pn)/(P+1);
    else
      proj_sigma = SX;

  return(proj_sigma);
}    


double H_Metropolis(struct ICMdata *ICMdata,double mm) 
{
  double new,std,var,beta,temp;
  double correction,dE;
  double CX;


  CX=ICMdata->CX; 
  
  var = 1/ICMdata->D2;
  if (var < 0.0)
    {	
      fprintf(stderr,"Second derivative of data model is negative!!\n");
      exit(-1);
    }
  std = sqrt(var);


  /* if mode of distrib. is >  -std/16.0, i.e. 'almost' positive */
  /* or if derivative of energy at x_s=0 <=0, then */
  /* use Gaussian transition probability */
  if ( mm > -std/16.0 || (temp=diff_cost(0.0,(void*)(ICMdata))) <= 0.0 )
    {
      /* generate x from normal dist N(mm,std) */
      /* accept only positive state x */
      while((new=std*normal() + mm)<0.0);
      /* correction = log(q(x',x)) - log(q(x,x')) */  
      correction = (CX-mm)*(CX-mm)/(2.0*var);
      correction -= (new-mm)*(new-mm)/(2.0*var);
    }
  else /* use exponential transition probability */
    {  /* to model tail of Gaussian with negative mean */
      /* this is done since want to generate positive x */
      beta = 1.0/temp;
      new = beta*exprand();
      correction = (CX-new)/beta;
    }
      

  /* Hastings-Metropolis equation */
  dE = cost(new,ICMdata) - cost(CX,ICMdata) + correction;
      
  /* accept or not accept new x */
  if (!(dE < 0.0 || (random2() < exp(-dE))))
    new = CX;

  return(new);

}
double H_Metropolis_new(struct ICMdata *ICMdata,double m,double *S,struct column *sp_col,struct all_data *data)
{
 
    double new,std,var,beta,temp;
 
  double mm;
  double CX;
  double dE,correction;


  mm=m;
  CX=ICMdata->CX;

  var = 1/ICMdata->D2;
  if (var < 0.0)
    {	
      fprintf(stderr,"Second derivative of data model is negative!!\n");
      exit(-1);
    }
  std = sqrt(var);


  /* if mode of distrib. is >  -std/16.0, i.e. 'almost' positive */
  /* or if derivative of energy at x_s=0 <=0, then */
  /* use Gaussian transition probability */
  if ( mm > -std/16.0 || (temp=diff_cost(0.0,(void*)(ICMdata))) <= 0.0 )
    {
      /* generate x from normal dist N(mm,std) */
      /* accept only positive state x */
      while((new=std*normal() + mm)<0.0);
      /* correction = log(q(x',x)) - log(q(x,x')) */  
      correction = (CX-mm)*(CX-mm)/(2.0*var);
      correction -= (new-mm)*(new-mm)/(2.0*var);
    }
  else /* use exponential transition probability */
    {  /* to model tail of Gaussian with negative mean */
      /* this is done since want to generate positive x */
      beta = 1.0/temp;
      new = beta*exprand();
      correction = (CX-new)/beta;
    }
      

  /* Hastings-Metropolis equation */
  /*dE = cost(new,ICMdata) - cost(CX,ICMdata) + correction;*/
  dE = delta_cost(new,ICMdata, sp_col, data, S) + correction;
  
  /* accept or not accept new x */
  if (!(dE < 0.0 || (random2() < exp(-dE))))
    new = CX;

  return(new);

}
double proj(double *x, double *y, int n)
{
  int i;
  double mx=0,my=0,mxy=0,mxsq=0,m,c;
  
  for(i=0;i<n;i++)
    {
      mx += x[i];
      my += y[i];
      mxy += x[i]*y[i];
      mxsq += x[i]*x[i];
    }
  mx /= n;
  my /= n;
  mxy /= n;
  mxsq /= n;
  
  m = (mxy - mx*my)/(mxsq - mx*mx);
  c = m*mx - my;
  
  return( c/m );
}

double delta_cost(double x,struct ICMdata *ICMdata, struct column *sp_col, struct all_data *data, double *S) 
{
  int i;
  double delta_x;
  double dllike=0.0,delta_cost,olike_i=0.0,nlike_i=0.0;
  double A_ij, y_i, S_i,yt_i;

  double delta_prior,prior=0.0,dprior=0.0;
  double temp;
 
 

  delta_x = x - ICMdata->CX;

  /* Compute delta_llike */
  for (i=0;i<sp_col->n_row;i++)
    {
      A_ij=sp_col->val[i];
      y_i=data->Y[0][sp_col->row[i]];
      S_i=S[sp_col->row[i]];

      yt_i=data->Yt[0][sp_col->row[i]];


      temp= yt_i*exp( -(S_i+A_ij*delta_x) );
      nlike_i = -temp + y_i*log( temp );

      temp= yt_i*exp( -S_i );
      olike_i = -temp + y_i*log( temp );
      
      dllike += (nlike_i-olike_i);
    }
   
  
 
  prior=(double)0.0;
  dprior=(double)0.0;

  for(i=0;i<4;i++)
    {
      prior -= pow(fabs(ICMdata->CX - ICMdata->neigh.NX[i]),ICMdata->P+1);
      prior += pow(fabs(x - ICMdata->neigh.NX[i]),ICMdata->P+1);
      dprior -= pow(fabs(ICMdata->CX - ICMdata->neigh.DX[i]),ICMdata->P+1);
      dprior += pow(fabs(x - ICMdata->neigh.DX[i]),ICMdata->P+1);
    }
  prior  = prior*WN;
  dprior  = dprior*WD;

  delta_prior = (prior+dprior)/( (ICMdata->P+1)*ICMdata->SX );


  delta_cost = -dllike + delta_prior;

  return(delta_cost);
}




