#include "fr_recon.h"




double relax_recon(struct all_data *data,double *S,struct geom_info *geom, double **profiles, struct cmdline_params *cmdline,struct column *sp_col,double P,double SX)
{
  
  int i,j,s,iter,N,M;
  double diff,new;
  double **X,**Y;
  double l_like;
  struct ICMdata ICMdata;
  



  X=data->X;
  Y=data->Y;
  N = geom->n_y;
  M = geom->n_x;
   
  /* ICM data */
  ICMdata.SX=SX;
  ICMdata.P=P;

  for(iter=0;iter<cmdline->no_iter_rec;iter++)
    {
      for (s=0;s<N*M;s++) /* loop over pixel locations */
	{
	  /* determine i,j image location, depending on iter go in*/
	  /* row or column order and also start from different corners */
	  imgcoord(&i,&j,s,iter,N,M);


	  if (data->recon_mask[i][j])
	    {
	      /* compute column of A matrix for loc. i,j */
	      A_comp(i,j,&data->sino,geom,profiles,sp_col);
	      
	      /* if more than 0 projections intersect location i,j */
	      if(sp_col->n_row)
		{
		  new=ICMstep(&ICMdata,data,S,i,j,geom,sp_col);
		  
		  /* In reconstruction mode, use ICM update , clip to >= 0.0 */
		  if( new < 0.0)
		    new = 0.0;
		  
		  X[i][j]=new;
		  		  
		  /* If pixel-value was changed, update S=Ax */
		  diff = X[i][j] - ICMdata.CX;
		  if(diff)
		    updateS(S,sp_col,diff);

	     
		} /* end if sp_col */
	    } /* end if recon_mask */
	} /* end for s */
      

   


      l_like=calc_llikelihood(data,geom, S, P, ICMdata.SX);
      printf("%d. Normalized log-posterior %f\n",iter+1,l_like/(N*M));



    } /* end for iter */


  return(ICMdata.SX);
}	




/* This calculates log posterior, not log likelihood */
double calc_llikelihood(struct all_data *data,struct geom_info *geom, double *S, double P, double SX)
{
  double l_like = 0.0;
  int i;
  int M,N;
  double temp,temp2;

  N = geom->n_y;
  M = geom->n_x;    
    
  for(i=0;i<data->sino.n_t*data->sino.n_theta;i++)
    {
      if (S[i] > 0.0)
	temp = data->Yt[0][i]*exp(-S[i]);
      else
	temp = data->Yt[0][i];

      temp2 = -temp + data->Y[0][i]*log( temp );
      l_like += temp2;
    }
   
  /* subtract log prior */
  l_like -= energy8_toroid_dbl(data->X,N,M,P+1)/((P+1)*SX);
  
  return(l_like);
}




